import os
import subprocess
import tempfile
import re
import io

from pyrpkg.errors import rpkgError
from pyparsing import Combine, CharsNotIn, Regex, ZeroOrMore, Group, Suppress, Literal

setup_path = os.path.join(
    os.path.dirname(__file__), 'preproc', 'setup')


def log_line_strip(line):
    return re.sub(r'(DEBUG|INFO|ERROR):\s*(.*)\s*', r'\2', line)


def invoke_wrapper(cwd, input_path, outdir, version_bump, env_store_name, app_log, break_on_failure):
    version_bump = str(version_bump) if version_bump else ''

    def invoke(tokens):
        retval = ""
        for token in tokens:
            cmd = ['/bin/bash', '-c' , '. {0}; {1};'.format(setup_path, token)]

            env = {
                'OUTDIR': outdir,
                'VERSION_BUMP': version_bump,
                'INPUT_PATH': input_path,
                'ENV_STORE': env_store_name,
            }
            p = subprocess.Popen(cmd, env=env, cwd=cwd,
                                 stderr=subprocess.PIPE,
                                 stdout=subprocess.PIPE)

            app_log.debug('> {0}'.format(token))
            app_log.debug('> cwd: {0}'.format(cwd))
            app_log.debug('> env: {0}'.format(env))

            (stdout, stderr) = p.communicate()

            for line in stderr.decode('utf-8').split('\n'):
                if not line:
                    continue

                if line.startswith('DEBUG:'):
                    log_func = app_log.debug
                elif line.startswith('INFO:'):
                    log_func = app_log.info
                else:
                    log_func = app_log.error

                stripped_line = log_line_strip(line)
                log_func(stripped_line)

            if p.returncode != 0:
                error_msg = "{0} failed with value {1}".format(
                    token, p.returncode)

                if break_on_failure:
                    raise rpkgError(error_msg)

                app_log.error(error_msg)

            app_log.debug('---')
            retval += '\n' if retval else ''
            retval += stdout.decode('utf-8')

        return retval
    return invoke


def get_parse_expr():
    opener, closer = '{{{', '}}}'
    nonquoted = ~Literal(opener) + ~Literal(closer) + CharsNotIn('\n', exact=1)
    quoted = Regex(r'".*"') ^ Regex(r"'.*'")
    content = Combine(ZeroOrMore(nonquoted^quoted)).setParseAction(lambda t:t[0].strip())
    return Suppress(opener) + content + Suppress(closer)


def preprocess(path, input_path, output_path, app_log, dump_spec_only=False, version_bump=False, break_on_failure=True):
    outdir = os.path.dirname(output_path) if not dump_spec_only else ''
    env_store_file = tempfile.NamedTemporaryFile(delete=False)

    with io.open(input_path, 'r', encoding='utf-8') as f:
        in_data = f.read()

        expr = get_parse_expr().setParseAction(invoke_wrapper(
            path, input_path, outdir, version_bump, env_store_file.name, app_log, break_on_failure))
        out_data = expr.transformString(in_data)

        outfile = io.open(output_path, 'w', encoding='utf-8')
        outfile.write(out_data)
        outfile.close()

    env_store_file.close()
    os.unlink(env_store_file.name)
